/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora's mesh-related element routines
 *
 *  Copyright (C) 2015 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#ifdef WASORA_MESH
#include <wasora.h>
#include "mesh.h"

/*

Line:                   Line3:           Line4:    
                                                
0----------1 --> u      0-----2----1     0----2----3----1

Triangle:               Triangle6:          Triangle9/10:          Triangle12/15:

v                                                              
^                                                                   2 
|                                                                   | \ 
2                       2                    2                      9   8
|`\                     |`\                  | \                    |     \ 
|  `\                   |  `\                7   6                 10 (14)  7
|    `\                 5    `4              |     \                |         \ 
|      `\               |      `\            8  (9)  5             11 (12) (13) 6
|        `\             |        `\          |         \            |             \
0----------1 --> u      0-----3----1         0---3---4---1          0---3---4---5---1

Quadrangle:            Quadrangle8:            Quadrangle9:

      v
      ^
      |
3-----------2          3-----6-----2           3-----6-----2 
|     |     |          |           |           |           | 
|     |     |          |           |           |           | 
|     +---- | --> u    7           5           7     8     5 
|           |          |           |           |           | 
|           |          |           |           |           | 
0-----------1          0-----4-----1           0-----4-----1 

Tetrahedron:                          Tetrahedron10:

                   v
                 .
               ,/
              /
           2                                     2                              
         ,/|`\                                 ,/|`\                          
       ,/  |  `\                             ,/  |  `\       
     ,/    '.   `\                         ,6    '.   `5     
   ,/       |     `\                     ,/       8     `\   
 ,/         |       `\                 ,/         |       `\ 
0-----------'.--------1 --> u         0--------4--'.--------1
 `\.         |      ,/                 `\.         |      ,/ 
    `\.      |    ,/                      `\.      |    ,9   
       `\.   '. ,/                           `7.   '. ,/     
          `\. |/                                `\. |/       
             `3                                    `3        
                `\.
                   ` w

Hexahedron:             Hexahedron20:          Hexahedron27:

       v
3----------2            3----13----2           3----13----2     
|\     ^   |\           |\         |\          |\         |\    
| \    |   | \          | 15       | 14        |15    24  | 14  
|  \   |   |  \         9  \       11 \        9  \ 20    11 \  
|   7------+---6        |   7----19+---6       |   7----19+---6 
|   |  +-- |-- | -> u   |   |      |   |       |22 |  26  | 23| 
0---+---\--1   |        0---+-8----1   |       0---+-8----1   | 
 \  |    \  \  |         \  17      \  18       \ 17    25 \  18
  \ |     \  \ |         10 |        12|        10 |  21    12| 
   \|      w  \|           \|         \|          \|         \| 
    4----------5            4----16----5           4----16----5 

Prism:                      Prism15:               Prism18:

           w
           ^
           |
           3                       3                      3        
         ,/|`\                   ,/|`\                  ,/|`\      
       ,/  |  `\               12  |  13              12  |  13    
     ,/    |    `\           ,/    |    `\          ,/    |    `\  
    4------+------5         4------14-----5        4------14-----5 
    |      |      |         |      8      |        |      8      | 
    |    ,/|`\    |         |      |      |        |    ,/|`\    | 
    |  ,/  |  `\  |         |      |      |        |  15  |  16  | 
    |,/    |    `\|         |      |      |        |,/    |    `\| 
   ,|      |      |\        10     |      11       10-----17-----11
 ,/ |      0      | `\      |      0      |        |      0      | 
u   |    ,/ `\    |    v    |    ,/ `\    |        |    ,/ `\    | 
    |  ,/     `\  |         |  ,6     `7  |        |  ,6     `7  | 
    |,/         `\|         |,/         `\|        |,/         `\| 
    1-------------2         1------9------2        1------9------2 

Pyramid:                     Pyramid13:                   Pyramid14:

               4                            4                            4
             ,/|\                         ,/|\                         ,/|\
           ,/ .'|\                      ,/ .'|\                      ,/ .'|\
         ,/   | | \                   ,/   | | \                   ,/   | | \
       ,/    .' | `.                ,/    .' | `.                ,/    .' | `.
     ,/      |  '.  \             ,7      |  12  \             ,7      |  12  \
   ,/       .' w |   \          ,/       .'   |   \          ,/       .'   |   \
 ,/         |  ^ |    \       ,/         9    |    11      ,/         9    |    11
0----------.'--|-3    `.     0--------6-.'----3    `.     0--------6-.'----3    `.
 `\        |   |  `\    \      `\        |      `\    \     `\        |      `\    \
   `\     .'   +----`\ - \ -> v  `5     .'        10   \      `5     .' 13     10   \
     `\   |    `\     `\  \        `\   |           `\  \       `\   |           `\  \ 
       `\.'      `\     `\`          `\.'             `\`         `\.'             `\` 
          1----------------2            1--------8-------2           1--------8-------2
                    `\
                       u
*/

int wasora_mesh_element_types_init(void) {

  element_type_t *element_type;
  int v;
  
  wasora_mesh.element_type = calloc(16, sizeof(element_type_t));

  // undefined
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_UNDEFINED];
  element_type->name = strdup("undefined");
  element_type->id = ELEMENT_TYPE_UNDEFINED;
  element_type->dim = 0;
  element_type->nodes = 0;
  element_type->faces = 0;
  element_type->nodes_per_face = 0;
  element_type->h = NULL;
  element_type->dhdr = NULL;
  element_type->point_in_element = NULL;
  element_type->element_volume = 0;
  
  // line
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_LINE];
  element_type->name = strdup("line");
  element_type->id = ELEMENT_TYPE_LINE;
  element_type->dim = 1;
  element_type->nodes = 2;
  element_type->faces = 2;
  element_type->nodes_per_face = 1;
  element_type->h = mesh_two_node_line_h;
  element_type->dhdr = mesh_two_node_line_dhdr;
  element_type->point_in_element = mesh_point_in_line;
  element_type->element_volume = mesh_line_vol;

  element_type->n_gauss = 2;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }
  
  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = -1.0/M_SQRT3;

  element_type->w_gauss[1] = 1.0;
  element_type->r_gauss[1][0] = +1.0/M_SQRT3;


  // triangle 
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_TRIANGLE];
  element_type->name = strdup("triangle");
  element_type->id = ELEMENT_TYPE_TRIANGLE;
  element_type->dim = 2;
  element_type->nodes = 3;
  element_type->faces = 3;
  element_type->nodes_per_face = 2;
  element_type->h = mesh_three_node_triang_h;
  element_type->dhdr = mesh_three_node_triang_dhdr;
  element_type->point_in_element = mesh_point_in_triangle;
  element_type->element_volume = mesh_triang_vol;

  element_type->n_gauss = 3;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }

  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = -1.0/M_SQRT3;
  element_type->r_gauss[0][1] = -1.0/M_SQRT3;
  
  element_type->w_gauss[1] = 1.0;
  element_type->r_gauss[1][0] = +1.0/M_SQRT3;
  element_type->r_gauss[1][1] = -1.0/M_SQRT3;
  
  element_type->w_gauss[2] = 2.0;
  element_type->r_gauss[2][0] = 0;
  element_type->r_gauss[2][1] = +1.0/M_SQRT3;
  
  
  // quadrangle
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_QUADRANGLE];
  element_type->name = strdup("quadrangle");
  element_type->id = ELEMENT_TYPE_QUADRANGLE;
  element_type->dim = 2;
  element_type->nodes = 4;
  element_type->faces = 4;
  element_type->nodes_per_face = 2;
  element_type->h = mesh_four_node_quad_h;
  element_type->dhdr = mesh_four_node_quad_dhdr;
  element_type->point_in_element = mesh_point_in_quadrangle;
  element_type->element_volume = mesh_quad_vol;

  element_type->n_gauss = 4;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }

  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = -1.0/M_SQRT3;
  element_type->r_gauss[0][1] = -1.0/M_SQRT3;

  element_type->w_gauss[1] = 1.0;
  element_type->r_gauss[1][0] = +1.0/M_SQRT3;
  element_type->r_gauss[1][1] = -1.0/M_SQRT3;

  element_type->w_gauss[2] = 1.0;
  element_type->r_gauss[2][0] = +1.0/M_SQRT3;
  element_type->r_gauss[2][1] = +1.0/M_SQRT3;

  element_type->w_gauss[3] = 1.0;
  element_type->r_gauss[3][0] = -1.0/M_SQRT3;
  element_type->r_gauss[3][1] = +1.0/M_SQRT3;
  
  // tetrahedron  
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_TETRAHEDRON];
  element_type->name = strdup("tetrahedron");
  element_type->id = ELEMENT_TYPE_TETRAHEDRON;
  element_type->dim = 3;
  element_type->nodes = 4;
  element_type->faces = 4;
  element_type->nodes_per_face = 3;
  element_type->h = mesh_four_node_tetrahedron_h;
  element_type->dhdr = mesh_four_node_tetrahedron_dhdr;
  element_type->point_in_element = mesh_point_in_tetrahedron;
  element_type->element_volume = mesh_tetrahedron_vol;
  
  element_type->n_gauss = 4;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }

  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = -1.0/M_SQRT3;
  element_type->r_gauss[0][1] = -1.0/M_SQRT3;
  element_type->r_gauss[0][2] = -1.0/M_SQRT3;

  element_type->w_gauss[1] = 1.0;
  element_type->r_gauss[1][0] = +1.0/M_SQRT3;
  element_type->r_gauss[1][1] = -1.0/M_SQRT3;
  element_type->r_gauss[1][2] = -1.0/M_SQRT3;

  element_type->w_gauss[2] = 2.0;
  element_type->r_gauss[2][0] = 0;
  element_type->r_gauss[2][1] = +1.0/M_SQRT3;
  element_type->r_gauss[2][2] = -1.0/M_SQRT3;

  element_type->w_gauss[3] = 4.0;
  element_type->r_gauss[3][0] = 0/M_SQRT3;
  element_type->r_gauss[3][1] = 0/M_SQRT3;
  element_type->r_gauss[3][2] = +1.0/M_SQRT3;

  
  // hexahedron  
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_HEXAHEDRON];
  element_type->name = strdup("hexahedron");
  element_type->id = ELEMENT_TYPE_HEXAHEDRON;
  element_type->dim = 3;
  element_type->nodes = 8;
  element_type->faces = 6;
  element_type->nodes_per_face = 4;
  element_type->h = mesh_eight_node_hexahedron_h;
  element_type->dhdr = mesh_eight_node_hexahedron_dhdr;
  element_type->point_in_element = mesh_point_in_hexahedron;
  element_type->element_volume = mesh_hexahedron_vol;

  element_type->n_gauss = 8;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }

  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = -1.0/M_SQRT3;
  element_type->r_gauss[0][1] = -1.0/M_SQRT3;
  element_type->r_gauss[0][2] = -1.0/M_SQRT3;

  element_type->w_gauss[1] = 1.0;
  element_type->r_gauss[1][0] = +1.0/M_SQRT3;
  element_type->r_gauss[1][1] = -1.0/M_SQRT3;
  element_type->r_gauss[1][2] = -1.0/M_SQRT3;

  element_type->w_gauss[2] = 1.0;
  element_type->r_gauss[2][0] = +1.0/M_SQRT3;
  element_type->r_gauss[2][1] = +1.0/M_SQRT3;
  element_type->r_gauss[2][2] = -1.0/M_SQRT3;

  element_type->w_gauss[3] = 1.0;
  element_type->r_gauss[3][0] = -1.0/M_SQRT3;
  element_type->r_gauss[3][1] = +1.0/M_SQRT3;
  element_type->r_gauss[3][2] = -1.0/M_SQRT3;

  element_type->w_gauss[4] = 1.0;
  element_type->r_gauss[4][0] = -1.0/M_SQRT3;
  element_type->r_gauss[4][1] = -1.0/M_SQRT3;
  element_type->r_gauss[4][2] = +1.0/M_SQRT3;

  element_type->w_gauss[5] = 1.0;
  element_type->r_gauss[5][0] = +1.0/M_SQRT3;
  element_type->r_gauss[5][1] = -1.0/M_SQRT3;
  element_type->r_gauss[5][2] = +1.0/M_SQRT3;

  element_type->w_gauss[6] = 1.0;
  element_type->r_gauss[6][0] = +1.0/M_SQRT3;
  element_type->r_gauss[6][1] = +1.0/M_SQRT3;
  element_type->r_gauss[6][2] = +1.0/M_SQRT3;

  element_type->w_gauss[7] = 1.0;
  element_type->r_gauss[7][0] = -1.0/M_SQRT3;
  element_type->r_gauss[7][1] = +1.0/M_SQRT3;
  element_type->r_gauss[7][2] = +1.0/M_SQRT3;
  
  
  // prism  
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_PRISM];
  element_type->name = strdup("prism");
  element_type->id = ELEMENT_TYPE_PRISM;
  element_type->dim = 3;
  element_type->nodes = 6;
  element_type->faces = 5;
  element_type->nodes_per_face = 4;   // Ojo aca que en nodos por cara pusimos el maximo valor (4) ya que depende de la cara
  element_type->h = mesh_six_node_prism_h;
  element_type->dhdr = mesh_six_node_prism_dhdr;
  element_type->point_in_element = mesh_point_in_prism;
  element_type->element_volume = mesh_prism_vol;

  element_type->n_gauss = 6;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }

  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = -1.0/M_SQRT3;
  element_type->r_gauss[0][1] = -1.0/M_SQRT3;
  element_type->r_gauss[0][2] = -1.0/M_SQRT3;
  
  element_type->w_gauss[1] = 1.0;
  element_type->r_gauss[1][0] = +1.0/M_SQRT3;
  element_type->r_gauss[1][1] = -1.0/M_SQRT3;
  element_type->r_gauss[1][2] = -1.0/M_SQRT3;
  
  element_type->w_gauss[2] = 2.0;
  element_type->r_gauss[2][0] = 0;
  element_type->r_gauss[2][1] = +1.0/M_SQRT3;
  element_type->r_gauss[2][2] = -1.0/M_SQRT3;
  
  element_type->w_gauss[3] = 1.0;
  element_type->r_gauss[3][0] = -1.0/M_SQRT3;
  element_type->r_gauss[3][1] = -1.0/M_SQRT3;
  element_type->r_gauss[3][2] = +1.0/M_SQRT3;
  
  element_type->w_gauss[4] = 1.0;
  element_type->r_gauss[4][0] = +1.0/M_SQRT3;
  element_type->r_gauss[4][1] = -1.0/M_SQRT3;
  element_type->r_gauss[4][2] = +1.0/M_SQRT3;
  
  element_type->w_gauss[5] = 2.0;
  element_type->r_gauss[5][0] = 0;
  element_type->r_gauss[5][1] = +1.0/M_SQRT3;
  element_type->r_gauss[5][2] = +1.0/M_SQRT3;

  // not supported  
  wasora_mesh.element_type[8].dim = -1;
  wasora_mesh.element_type[9].dim = -1;
  wasora_mesh.element_type[10].dim = -1;
  wasora_mesh.element_type[11].dim = -1;
  wasora_mesh.element_type[12].dim = -1;
  wasora_mesh.element_type[13].dim = -1;
  wasora_mesh.element_type[14].dim = -1;

  
  // point  
  element_type = &wasora_mesh.element_type[ELEMENT_TYPE_POINT];
  element_type->name = strdup("point");
  element_type->id = ELEMENT_TYPE_POINT;
  element_type->dim = 0;
  element_type->nodes = 1;
  element_type->faces = 0;
  element_type->nodes_per_face = 0;
  element_type->h = mesh_one_node_point_h;
  element_type->dhdr = mesh_one_node_point_dhdr;
  element_type->element_volume = mesh_point_vol;
  element_type->point_in_element = NULL;

  element_type->n_gauss = 1;
  element_type->w_gauss = malloc(element_type->n_gauss * sizeof(double));
  element_type->r_gauss = malloc(element_type->n_gauss * sizeof(double *));
  for (v = 0; v < element_type->n_gauss; v++) {
    element_type->r_gauss[v] = malloc(element_type->dim * sizeof(double));
  }
  
  element_type->w_gauss[0] = 1.0;
  element_type->r_gauss[0][0] = 0;
  
  return WASORA_RUNTIME_OK;
  
};


// esta no rellena los nodos!
int mesh_create_element(element_t *element, int id, int type, physical_entity_t *physical_entity) {
 
  element->id = id;
  element->type = &(wasora_mesh.element_type[type]);
  element->node  = calloc(element->type->nodes, sizeof(node_t *));
  element->physical_entity = physical_entity;
  
  return WASORA_RUNTIME_OK;
}

int mesh_add_element_to_list(element_list_item_t **list, element_t *element) {
  
  element_list_item_t *item = calloc(1, sizeof(element_list_item_t));
  item->element = element;
  LL_APPEND(*list, item);
  
  return WASORA_RUNTIME_OK;
  
}

int mesh_compute_element_barycenter(element_t *element, double barycenter[]) {
  
  int j;
  
  barycenter[0] = barycenter[1] = barycenter[2] = 0;
  for (j = 0; j < element->type->nodes; j++) {
    barycenter[0] += element->node[j]->x[0];
    barycenter[1] += element->node[j]->x[1];
    barycenter[2] += element->node[j]->x[2];
  }
  barycenter[0] /= element->type->nodes; 
  barycenter[1] /= element->type->nodes; 
  barycenter[2] /= element->type->nodes; 
  
  return WASORA_RUNTIME_OK;
}


int mesh_node_indexes(mesh_t *mesh, int dofs) {
  int index = 0;
  int j, g;
  
  mesh->degrees_of_freedom = dofs;
  
  for (j = 0; j < mesh->n_nodes; j++) {
    mesh->node[j].index = malloc(mesh->degrees_of_freedom*sizeof(int));
    for (g = 0; g < mesh->degrees_of_freedom; g++) {
      mesh->node[j].index[g] = index++;
    }
  }
  
  return WASORA_RUNTIME_OK;
}
#endif
